local astrolabe = DongleStub("Astrolabe-1.0")
local event = {}

local currencies = {
	384, -- Dwarf
	398, -- Draenei
	393, -- Fossil
	394, -- Night Elf
	400, -- Nerubian
	397, -- Orc
	401, -- Tol'vir
	385, -- Troll
	399, -- Vrykul
}

-- assertTable(tbl, ...)
-- Checks that a table has a given field, without throwing an error.
local function assertTable(tbl, ...)
	if not (type(tbl) == "table") then return false end
	if ... then
		if tbl[...] then
			if select(2, ...) then
				return assertTable(tbl[...], select(2, ...))
			end
		else
			return false
		end
	end
	return true
end

local db = {}
db.show = true
db.track = true

------------------------------
-- Database version control --
------------------------------

-- Note: Release versions are integers. Beta versions are decimal points.

local version = 2
-- This is the database version number, NOT the file or addon version. It's used to determine when a
--	database update is needed.
local function defaultDB()
	return {
		show = true,
		track = true,
		version = version,
	}
end

local function updateDB(DB)
	if not DB.version then
		return defaultDB()
	end
	if DB.version == version then
		return DB
	end
	return defaultDB()
end

-------------------------
-- Icon object manager --
-------------------------
-- (to avoid creating frames unnecessarily)

local iconPool = {}
iconPool.numIcons = 0
local shownIcons = {}

-- getIcon()
-- Returns an unused icon from the icon pool.
-- If there are no unused icons, create a new icon and return it.
local function getIcon()
	local r
	if #iconPool > 0 then
		-- Remove the icon from the pool, add it to the display, and return it
		r = iconPool[#iconPool]
		iconPool[#iconPool] = nil
		shownIcons[r] = true
		return r
	else
		-- Create a new icon, add it to the display, and return it
		iconPool.numIcons = iconPool.numIcons + 1
		r = CreateFrame("Button", "ExcavatinatorMinimapButton" .. iconPool.numIcons, UIParent, "ExcavatinatorMinimapButtonTemplate")
		shownIcons[r] = true
		return r
	end
end

-- removeIcon(icon)
-- Removes an icon from the minimap, and returns it to the icon pool.
local function removeIcon(icon)
	if shownIcons[icon] then
		astrolabe:RemoveIconFromMinimap(icon)
		iconPool[#iconPool + 1] = icon
		shownIcons[icon] = nil
	end
end

-- removeAllIcons()
-- Removes all icons from the minimap and returns them to the icon pool
local function removeAllIcons()
	for k, v in pairs(shownIcons) do
		astrolabe:RemoveIconFromMinimap(k)
		iconPool[#iconPool + 1] = k
		shownIcons[k] = nil
	end
end

-----------------------
-- Display functions --
-----------------------

-- updateIcons()
-- Shows icons within a 250 yard radius; hides all others
local function updateIcons()
	if db.show then
		local c, z, x, y = astrolabe:GetCurrentPlayerPosition()
		if db[c] then
			if db[c][z] then
				for i=1, #db[c][z] do
					local distance = astrolabe:ComputeDistance(c, z, x, y, c, z, db[c][z][i].x, db[c][z][i].y)
					if distance < 250 then
						if not assertTable(db[c][z][i], "icon", 0) then
							local icon = getIcon()
							astrolabe:PlaceIconOnMinimap(icon, c, z, db[c][z][i].x, db[c][z][i].y)
							db[c][z][i].icon = icon
						end
					else
						removeIcon(db[c][z][i].icon)
						db[c][z][i].icon = nil
					end
				end
			end
		end
	end
end

----------------------
-- Database manager --
----------------------

-- addPosition(currency, optc, optz, optx, opty)
-- Adds a gather position to the database.
-- Positions that are within 10 yards of each other are considered the same position.
local function addPosition(currency, optc, optz, optx, opty)
	if db.track then
		-- Add or update the position
		local c, z, x, y
		if (optc and optz and optx and opty) then
			c, z, x, y = optc, optz, optx, opty
		else
			c, z, x, y = astrolabe:GetCurrentPlayerPosition()
		end
		if not db[c] then db[c] = {} end -- If no entry for the current continent, create one
		if not db[c][z] then db[c][z] = {} end -- If no entry for the current zone, create one
		local found = 0
		for i=1, #db[c][z] do
			local distance = astrolabe:ComputeDistance(c, z, x, y, c, z, db[c][z][i].x, db[c][z][i].y)
			if distance < 10 then
				-- It's an old position; add one artifact found
				db[c][z][i].gathers = db[c][z][i].gathers + 1
				if currency > 0 then db[c][z][i].race = currency end -- In case of importing, don't overwrite the race
				found = i
				break
			end
		end
		if found == 0 then
			-- It's a new position; add it to the database and make an icon for the position (if it's in the zone the player is actually in)
			db[c][z][#db[c][z] + 1] = { x = x, y = y, gathers = 1, race = currency }
			local cr, zr = astrolabe:GetCurrentPlayerPosition()
			if c == cr and z == zr then
				local icon = getIcon()
				astrolabe:PlaceIconOnMinimap(icon, c, z, x, y)
				db[c][z][#db[c][z]].icon = icon
			end
		end
	end
end

-- removePosition()
-- Removes the position the player is currently at from the database.
-- Serves no purpose in the addon proper; retained for testing purposes.
local function removePosition()
	local c, z, x, y = astrolabe:GetCurrentPlayerPosition()
	for i=1, #db[c][z] do
		local distance = astrolabe:ComputeDistance(c, z, x, y, c, z, db[c][z][i].x, db[c][z][i].y)
		if distance < 10 then
			table.remove(db[c][z], i)
		end
	end
end

local function importFromArchy()
	if not ArchyDB then
		print("Excavatinator: Cannot load from Archy: No Archy database found.")
	else
		if assertTable(ArchyDB, "global", "surveyNodes") then
			for k, v in pairs(ArchyDB.global.surveyNodes) do
				for i=1, #v do
					addPosition(0, v[i].m, v[i].f, v[i].x, v[i].y)
				end
			end
			removeAllIcons()
			updateIcons()
		end
	end
end

local function resetDB()
	local show, track = db.show, db.track
	table.wipe(db)
	db.show = show
	db.track = track
	removeAllIcons()
end

---------------------
-- Event functions --
---------------------

function event.ADDON_LOADED(addonName)
	if addonName == "Excavatinator" then
		if ExcavatinatorNodeDB then
			db = updateDB(ExcavatinatorNodeDB)
		end
	end
end

function event.PLAYER_LOGIN()
	if type(currencies[1]) == "number" then
		for i=1, #currencies do
			currencies[i] = GetCurrencyInfo(currencies[i])
		end
	end
end

function event.PLAYER_LEAVING_WORLD()
	ExcavatinatorNodeDB = db
	ExcavatinatorNodeDB.version = version
end

function event.CHAT_MSG_CURRENCY(msg)
	for i=1, #currencies do
		if strfind(msg, currencies[i], 0, true) then
			addPosition(i)
		end
	end
end

function event.ZONE_CHANGED_NEW_AREA()
	-- If the map is up when changing zones, icons may not be removed properly.
	-- Fix this by removing all icons one when the zone is changed.
	removeAllIcons()
end

-------------------
-- Event manager --
-------------------

local function onEvent(s, e, ...)
	if event[e] then
		event[e](...)
	end
end

local eventFrame = CreateFrame("Frame")
eventFrame:SetScript("OnEvent", onEvent)
for k, v in pairs(event) do
	eventFrame:RegisterEvent(k)
end

----------------------
-- OnUpdate manager --
----------------------

local passed, interval = 0, 1

local function onUpdate(_, elapsed)
	passed = passed + elapsed
	if passed > interval then
		updateIcons()
		passed = 0
	end
end

local updateFrame = CreateFrame("Frame")
updateFrame:SetScript("OnUpdate", onUpdate)

-------------------------
-- Confirmation frames --
-------------------------

StaticPopupDialogs["EXCAVATINATOR_DELETEDB"] = {
	text = "Are you sure you want to clear your node database? This can NOT be undone.",
	button1 = "Yes",
	button2 = "No",
	OnAccept = resetDB,
	timeout = 0,
	whileDead = true,
	hideOnEscape = true,
}

StaticPopupDialogs["EXCAVATINATOR_IMPORT_ARCHY"] = {
	text = "Import node data from Archy?",
	button1 = "Yes",
	button2 = "No",
	OnAccept = importFromArchy,
	timeout = 0,
	whileDead = true,
	hideOnEscape = true,
}

------------------------------------
-- Communicate with the main file --
------------------------------------
-- (this also adds to the API)

-- Excavatinator:SetMinimapDisplay(isOn)
-- Turns minimap icon display on or off
function Excavatinator:SetMinimapDisplay(isOn)
	db.show = isOn
	if not isOn then removeAllIcons() end
	updateIcons()
end

-- Excavatinator:GetMinimapDisplay()
-- Returns whether minimap display is on or not
function Excavatinator:GetMinimapDisplay()
	return db.show
end

-- Excavatinator:SetMinimapTracking(isOn)
-- Turns dig tracking on or off
function Excavatinator:SetMinimapTracking(isOn)
	db.track = isOn
end

-- Excavatinator:GetMinimapTracking()
-- Returns whether minimap tracking is on or off
function Excavatinator:GetMinimapTracking()
	return db.track
end

-- Excavatinator:ResetDB()
-- Resets Excavatinator's node database. The user will have to confirm this.
function Excavatinator:ResetDB()
	StaticPopup_Show("EXCAVATINATOR_DELETEDB")
end

-- Excavatinator:ImportDatabase(addonID)
-- Imports the database of a supported addon.
function Excavatinator:ImportDatabase(addonID)
	if addonID == 1 then
		StaticPopup_Show("EXCAVATINATOR_IMPORT_ARCHY")
	end
end